defmodule Swoosh.InMemoryMailbox do
  @moduledoc ~S"""
  In-memory mailbox used by the
  [Swoosh.Adapters.Local](Swoosh.Adapters.Local.html) module.

  The emails in this mailbox are stored in memory and won't persist once your
  application is stopped.
  """

  use GenServer

  @doc """
  Starts the InMemoryMailbox server
  """
  def start_link() do
    GenServer.start_link(__MODULE__, [], name: __MODULE__)
  end

  @doc """
  Stops the InMemoryMailbox server
  """
  def stop() do
    GenServer.stop(__MODULE__)
  end

  @doc ~S"""
  Push a new email into the mailbox.

  In order to make it easy to fetch a single email, a `Message-ID` header is
  added to the email before being stored.

  ## Examples

      iex> email = %Swoosh.Email{} |> from("tony@stark.com")
      %Swoosh.Email{from: {"", "tony@stark.com"}, [...]}
      iex> InMemoryMailbox.push(email)
      %Swoosh.Email{from: {"", "tony@stark.com"}, headers: %{"Message-ID": "A1B2C3"}, [...]}
  """
  def push(email) do
    GenServer.call(__MODULE__, {:push, email})
  end

  @doc ~S"""
  Pop the last email from the mailbox.

  ## Examples

      iex> email = %Swoosh.Email{} |> from("tony@stark.com")
      %Swoosh.Email{from: {"", "tony@stark.com"}, [...]}
      iex> InMemoryMailbox.push(email)
      %Swoosh.Email{from: {"", "tony@stark.com"}, headers: %{"Message-ID": "A1B2C3"}, [...]}
      iex> InMemoryMailbox.all() |> Enum.count()
      1
      iex> InMemoryMailbox.pop()
      %Swoosh.Email{from: {"", "tony@stark.com"}, headers: %{"Message-ID": "A1B2C3"}, [...]}
      iex> InMemoryMailbox.all() |> Enun.count()
      0
  """
  def pop() do
    GenServer.call(__MODULE__, :pop)
  end

  @doc ~S"""
  Get a specific email from the mailbox.

  ## Examples

      iex> email = %Swoosh.Email{} |> from("tony@stark.com")
      %Swoosh.Email{from: {"", "tony@stark.com"}, [...]}
      iex> InMemoryMailbox.push(email)
      %Swoosh.Email{from: {"", "tony@stark.com"}, headers: %{"Message-ID": "A1B2C3"}, [...]}
      iex> InMemoryMailbox.get("A1B2C3")
      %Swoosh.Email{from: {"", "tony@stark.com"}, headers: %{"Message-ID": "A1B2C3"}, [...]}
  """
  def get(id) do
    GenServer.call(__MODULE__, {:get, id})
  end

  @doc ~S"""
  List all the emails in the mailbox.

  ## Examples

      iex> email = %Swoosh.Email{} |> from("tony@stark.com")
      %Swoosh.Email{from: {"", "tony@stark.com"}, [...]}
      iex> InMemoryMailbox.push(email)
      %Swoosh.Email{from: {"", "tony@stark.com"}, headers: %{"Message-ID": "A1B2C3"}, [...]}
      iex> InMemoryMailbox.all()
      [%Swoosh.Email{from: {"", "tony@stark.com"}, headers: %{"Message-ID": "A1B2C3"}, [...]}]
  """
  def all() do
    GenServer.call(__MODULE__, :all)
  end

  @doc ~S"""
  Delete all the emails currently in the mailbox.

  ## Examples

      iex> email = %Swoosh.Email{} |> from("tony@stark.com")
      %Swoosh.Email{from: {"", "tony@stark.com"}, [...]}
      iex> InMemoryMailbox.push(email)
      %Swoosh.Email{from: {"", "tony@stark.com"}, headers: %{"Message-ID": "A1B2C3"}, [...]}
      iex> InMemoryMailbox.delete_all()
      :ok
      iex> InMemoryMailbox.list()
      []
  """
  def delete_all() do
    GenServer.call(__MODULE__, :delete_all)
  end

  # Callbacks

  def init(_args) do
    {:ok, []}
  end

  def handle_call({:push, email}, _from, state) do
    id = :crypto.rand_bytes(16) |> Base.encode16
    email = email |> Swoosh.Email.header("Message-ID", id)
    {:reply, email, [email] ++ state}
  end

  def handle_call(:pop, _from, [h|t]) do
    {:reply, h, t}
  end

  def handle_call({:get, id}, _from, state) do
    email = Enum.find(state, nil, fn %Swoosh.Email{headers: %{"Message-ID" => mid}} -> mid == id end)
    {:reply, email, state}
  end

  def handle_call(:all, _from, state) do
    {:reply, state, state}
  end

  def handle_call(:delete_all, _from, _state) do
    {:reply, :ok, []}
  end
end
